#ifndef opl3emu_operator_hh
#define opl3emu_operator_hh

class Operator
{
public:
	enum chaintype { mod_none, mod_freq, mod_add };

private:
	bool tremolo;              // 3.7 Hz. Depth controlled by system. 1 dB or 4.8 dB.
	bool vibrato;              // 6.1 Hz. Depth controlled by system. 7 or 14 cents.
	bool dosustain;
	bool KSR;                  // Decrease note lengths when freq grows?
	unsigned char freqmul;     // 0..15. Multiplies the frequency by
	                           //   0=0.5, 11=10, 13=12, 14=15, else freqmul
	
	unsigned char KSL;         // 0..3 - Attenuation.
	                           //   0 = no attenuation by octave
	                           //   1 = 1.5 dB / octave
	                           //   2 = 3.0 dB / octave
	                           //   3 = 6.0 dB / octave
	unsigned char attack;      // 0..15- Rising time.
	unsigned char decay;       // 0..15- 
	unsigned char sustain;     // 0..15- sustain*3 = Sustain level.
	                           //     When b1111, according to YMF262 documentation,
	                           //                 sustain level = 93 dB. (???)
	
	unsigned char release;     // 0..15- Rate of disappear after key-off.
	
	double attenuation;        // The one set by Write40(). Unit: dB.
	
	double effvol, outputlevel;
	double sustainlevel;
	double ownoutput;
	
	class timer *envelopetimer;
	unsigned envpos1, envpos2;
	
	Operator *target, *source;
	
	void FixEnvelope();
	bool EnvelopeTick();
	
	chaintype chainway;
	
public:

	Operator();
	~Operator();
	
	double multiplyfreq(double freq) const;
	
	void Write20(unsigned char value);
	void Write40(unsigned char value);
	void Write60(unsigned char value);
	void Write80(unsigned char value);
	
	void Modulate(Operator &o);
	void AddTo(Operator &o);
	void Unlink(void);
	
	void Run(double secs);
	
	double lastoutput() const { return ownoutput; }	
	double Produce(unsigned waveno, double freq, double angle, double shift=0.0);
	
	bool IsProductive() const;
	chaintype SourceRelation(const Operator &Source) const;
	chaintype TargetRelation() const;

	void KeyOn();
	void KeyOff();	
};

extern unsigned OperatorPortOffset(unsigned channel);
extern unsigned PortOffsetOperator(unsigned offset);

#endif
